import yaml
import os
import re

def parse_yaml_header(file):
    lines = []
    line = file.readline()
    while line != "---" + os.linesep:
        lines.append(line)
        line = file.readline()

    return yaml.safe_load("".join(lines))

PUNCTUATION = ["#", "$", "!", ".", ",", "?", "/", ":", ";", "`", " ", "-", "+", "=", os.linesep] + [str(i) for i in range(0, 10)]


def get_tags_from_line(line) -> [str]:
    pos_tags = [i for i, char in enumerate(line) if char == '#']
    tags = []
    for i in pos_tags:
        if i == 0 or line[i - 1] == ' ':
            index = next((index for index, c in enumerate(line[i+1:]) if c in PUNCTUATION), -1)
            if index == -1:
                tags.append(line[i+1:])
            else:
                tag = line[i + 1:index + i + 1]
                if len(tag) > 0:
                    tags.append(tag)
    return tags

def get_wikilinks_from_line(line) -> [str]:
    result = re.findall('\[\[(.*?)\]\]', line)
    if result:
        r = []
        for wikilink in result:
            r.append(wikilink.split("|")[0])
        return r
    return []

def escape_quotes(string) -> str:
    r1 = string.replace("\\\'", "\\\\\'")
    r1 = r1.replace("'", "\\\'")
    return r1.replace('\"', "\\\"")



from smdc.format import Format
from smdc.note import Note, Relationship
import io
import os
from smdc.format.util import parse_yaml_header, get_tags_from_line, get_wikilinks_from_line, PUNCTUATION


class TypedList(Format):

    def parse_word(self, line, index, breaks=[' ', os.linesep, ',']):
        if index >= len(line):
            return index
        for j in range(index, len(line)):
            if line[j] in breaks:
                return j

        return j

    def move_index(self, line, index):
        if index >= len(line):
            return index
        for j in range(index, len(line)):
            if line[j] != ' ':
                return j

        return j

    def parse(self, file: io.TextIOWrapper, name, parsed_notes: [Note]) -> Note:
        line = file.readline()
        parsed_yaml = None
        # Find YAML header, or continue
        while line:
            if line.strip():
                if line == '---' + os.linesep:
                    try:
                        parsed_yaml = parse_yaml_header(file)
                    except Exception as e:
                        print(e)
                break
            line = file.readline()

        content = []
        relations = {}
        tags = set()
        while line:
            if line.startswith("- ") and len(line) > 2:
                is_rel = True
                index = self.parse_word(line, 2, breaks=PUNCTUATION)
                type = line[2:index]
                if len(type) != 0:
                    index = self.move_index(line, index + 1)
                    words = []
                    while index < len(line) - 2:
                        new_index = self.parse_word(line, index)
                        words.append(line[index:new_index])
                        index = self.move_index(line, new_index + 1)
                    year = None
                    trgts = []
                    active_trgt = None
                    for i, word in enumerate(words):
                        if word[:2] == "[[":
                            if active_trgt:
                                is_rel = False
                                break
                            if word[-2:] == "]]":
                                trgts.append(word[2:-2].split("|")[0])
                            else:
                                active_trgt = word[2:]
                            continue
                        elif word[-2:] == "]]":
                            if not active_trgt:
                                is_rel = False
                                break
                            trgts.append((active_trgt + " " +  word[:-2]).split("|")[0])
                            active_trgt = None
                            continue
                        if i == 0 and type in ['publishedIn', 'at']:
                            year = word
                        elif active_trgt:
                            active_trgt += " " + word
                        else:
                            is_rel = False
                            break
                    if is_rel:
                        for trgt in trgts:
                            properties = {}
                            if year:
                                properties["year"] = year
                            rel = Relationship(type, properties)
                            if trgt in relations:
                                relations[trgt].append(rel)
                            else:
                                relations[trgt] = [rel]
                        line = file.readline()
                        continue
            content.append(line)
            tags.update(get_tags_from_line(line))
            # TODO: Save aliases as Relation property
            for wikilink in get_wikilinks_from_line(line):
                rel = Relationship("inline", properties={"context": line})
                if wikilink in relations:
                    relations[wikilink].append(rel)
                else:
                    relations[wikilink] = [rel]
            line = file.readline()

        return Note(name, tags, "".join(content), out_rels=relations, properties=parsed_yaml if parsed_yaml else {})

    def write(self, file, parsed_notes: [Note]):
        raise NotImplementedError


from smdc.format import Format
from smdc.note import Note, Relationship
import io
import os
from smdc.format.util import escape_quotes


def escape_cypher(string):
    r = escape_quotes(string)
    # Note: CYPHER doesn't allow putting semicolons in text, for some reason. This is lossy!
    r = r.replace(";", ",")
    r = r.replace("\\u", "\\\\u")
    if r and r[-1] == '\\':
        r += ' '
    return r

def to_cyper(parsed_notes: [Note]):
    lines = []
    # First create all nodes in the graph before doing the relationships, so they all exist.
    for name, note in parsed_notes.items():
        line = "CREATE ("
        if note.tags:
            line += ":" + ":".join(note.tags)
        line += " { name: '" + escape_cypher(name) + "', content: '" + escape_cypher(note.content) + "'"
        for property, value in note.properties.items():
            line += ", " + property + ": '" + escape_cypher(str(value)) + "'"
        line += "});"
        lines.append(line)

    notes_in_cypher = set(parsed_notes.keys())

    for name, note in parsed_notes.items():
        if not note.out_rels.keys():
            continue
        match_a = "MATCH (a)" + os.linesep + "WHERE a.name = '" + escape_cypher(name) + "'" + os.linesep
        for trgt, rels in note.out_rels.items():
            if trgt not in notes_in_cypher:
                lines.append("CREATE ({name:'" + escape_cypher(trgt) + "'});")
                notes_in_cypher.add(trgt)
            match_b = "MATCH (b)" + os.linesep + "WHERE b.name = '" + \
                      escape_cypher(trgt) + "'" + os.linesep
            for rel in rels:
                line = "CREATE (a)-[:" + rel.type + " {"
                properties = []
                for property, value in rel.properties.items():
                    properties.append(property + ": '" + escape_cypher(str(value)) + "'")
                line += ", ".join(properties)
                line += "}]->(b);"
                lines.append(match_a + match_b + line)
    return lines

class Cypher(Format):

    def parse(self, file: io.TextIOWrapper, name, parsed_notes: [Note]) -> Note:
        raise NotImplementedError

    def write(self, parsed_notes: [Note], args):
        lines = to_cyper(parsed_notes)
        with open(args.output + '.cypher', 'w', encoding='utf-8') as f:
            f.writelines(os.linesep.join(lines))


